from kubr.backends.volcano import VolcanoBackend
from kubr.commands.base import BaseCommand


class RmCommand(BaseCommand):
    @staticmethod
    def add_parser(subparsers, completer):
        rm_parser = subparsers.add_parser('rm', help='Delete a job')
        rm_parser.add_argument('job', help='Name of job to delete').completer = completer
        rm_parser.add_argument('-n', '--namespace', help='Namespace to delete job from', default='default')
        return rm_parser

    def __call__(self, job_name: str, namespace: str = 'default'):
        return self.backend.delete_job(job_name=job_name, namespace=namespace)



from itertools import zip_longest


def join_tables_horizontally(table1, table2, separator=' ' * 4):
    t1_rows = table1.split('\n')
    t2_rows = table2.split('\n')

    max_rows = max(len(t1_rows), len(t2_rows))
    t1_max_width = max([len(row) for row in t1_rows])
    t2_max_width = max([len(row) for row in t2_rows])

    output = []
    for t1_row, t2_row in zip_longest(t1_rows, t2_rows, fillvalue=None):
        if t1_row is not None and t2_row is not None:
            output.append(f'{t1_row}{separator}{t2_row}')
        elif t1_row is None:
            output.append(f'{" " * t1_max_width}{separator}{t2_row}')
        elif t2_row is None:
            output.append(f'{t1_row}{separator}{" " * t2_max_width}')
    return '\n'.join(output)



from typing import Optional

from kubr.backends.volcano import VolcanoBackend
from kubr.commands.base import BaseCommand
from kubr.config.runner import RunnerConfig
from pydantic_yaml import parse_yaml_raw_as

class RunCommand(BaseCommand):
    @staticmethod
    def add_parser(subparsers):
        run_parser = subparsers.add_parser('run', help='Submit a new job')
        run_parser.add_argument('config', help='Path to run config', type=str)
        run_parser.add_argument('-i', '--image', help='Image to run')
        run_parser.add_argument('-e', '--entrypoint', help='Entrypoint to run')
        run_parser.add_argument('-n', '--namespace', help='Namespace to submit job to')
        run_parser.add_argument('--name', help='Name of job')
        run_parser.add_argument('-it', '--interactive', help='Run job interactively', action='store_true',
                                default=False)

    def __call__(self, config: str, image: Optional[str] = None,
                 name: Optional[str] = None, entrypoint: Optional[str] = None,
                 namespace: Optional[str] = None):
        with open(config, 'r') as f:
            config = f.read()
        config = parse_yaml_raw_as(RunnerConfig, config)

        config.experiment.name = name or config.experiment.name
        config.container.image = image or config.container.image
        config.container.entrypoint = entrypoint or config.container.entrypoint
        config.experiment.namespace = namespace or config.experiment.namespace

        return self.backend.run_job(config)



# PYTHON_ARGCOMPLETE_OK
import argcomplete
import argparse

from kubr.backends.volcano import VolcanoBackend
from kubr.commands.desc import DescribeCommand
from kubr.commands.logs import LogsCommand
from kubr.commands.ls import LsCommand
from kubr.commands.rm import RmCommand
from kubr.commands.run import RunCommand


def main():
    # TODO fix package install adding eval in bash\zsh for autocomplete
    # adding eval "$(register-python-argcomplete kubr)" to .bashrc or .zshrc
    # TODO fix autopilot deployment in GKE autopilot
    # TODO fix Volcano priority class in GKE (https://github.com/volcano-sh/volcano/issues/2379)

    backend = VolcanoBackend()
    arg = argparse.ArgumentParser(description='Kubr', add_help=True)
    arg.add_argument('--version', help='Get version of Kubr')
    subparsers = arg.add_subparsers(help='Commands', dest='command')

    run_parser = RunCommand.add_parser(subparsers)
    ls_parser = LsCommand.add_parser(subparsers)
    rm_parser = RmCommand.add_parser(subparsers, completer=backend._completion_list_running_jobs)
    desc_parser = DescribeCommand.add_parser(subparsers, completer=backend._completion_list_running_jobs)
    logs_parser = LogsCommand.add_parser(subparsers, completer=backend._completion_list_running_jobs)

    # attach_parser = AttachCommand.add_parser(subparsers, completer=backend._completion_list_running_jobs)
    # stat_parser = StatCommand.add_parser(subparsers, completer=backend._completion_list_running_jobs)
    # test_parser = TestCommand.add_parser(subparsers)

    argcomplete.autocomplete(arg)
    args = arg.parse_args()

    if args.command == 'run':
        operator = RunCommand()
        print(operator(config=args.config, image=args.image, entrypoint=args.entrypoint, namespace=args.namespace,
                       name=args.name))
    elif args.command == 'ls':
        operator = LsCommand()
        operator(namespace=args.namespace, show_all=args.all, head=args.top)
    elif args.command == 'rm':
        operator = RmCommand()
        print(operator(job_name=args.job, namespace=args.namespace))
    elif args.command == 'desc':
        operator = DescribeCommand()
        print(operator(job_name=args.job_name, namespace=args.namespace))
    elif args.command == 'logs':
        operator = LogsCommand()
        print(operator(job_name=args.job, namespace=args.namespace, tail=args.tail))
    elif args.command == 'attach':
        raise NotImplementedError  # TODO implement attach command
    elif args.command == 'stat':
        raise NotImplementedError  # TODO implement stat command
    elif args.command == 'test':
        raise NotImplementedError  # TODO implement test command -- run IB\scheduler\metrics\registry\ethernet tests
    else:
        arg.print_help()


if __name__ == '__main__':
    main()



from kubr.backends.volcano import VolcanoBackend
from typing import List

from kubr.commands.base import BaseCommand
from kubr.config.job import Job
from collections import defaultdict
import humanize
from tabulate import tabulate
from typing import Optional
from kubr.backends.utils import join_tables_horizontally
from datetime import datetime


def visualize_jobs(jobs: List[Job], head: Optional[int] = 10, show_all: bool = False):
    extracted_jobs = defaultdict(list)
    for job in jobs:
        if job.state in ['Pending', 'Running', 'Failed', 'Completed']:
            extracted_jobs[job.state].append(job)
        else:
            extracted_jobs['extra'].append(job)

    for state in ['Pending', 'Running', 'Completed', 'Failed', 'extra']:
        extracted_jobs[state].sort(key=lambda x: x.age, reverse=True)
        for job in extracted_jobs[state]:
            job.age = humanize.naturaldelta(datetime.utcnow() - job.age)
        if head:
            extracted_jobs[state] = extracted_jobs[state][:head]

        # TODO [ls] add pretty formatting that will show only first 10 jobs in completed and failed states are shown
        if not show_all and state in ['Completed', 'Failed']:
            extracted_jobs[state] = extracted_jobs[state][:5]

        dict_jobs = [job.dict() for job in extracted_jobs[state]]
        if len(dict_jobs):
            extracted_jobs[state] = tabulate(dict_jobs, headers='keys', tablefmt='grid')
        else:
            extracted_jobs[state] = ''

    # TODO pretty handling of empty list in running jobs
    result = join_tables_horizontally(extracted_jobs['Running'], extracted_jobs['Pending'])
    result += '\n\n' if len(result) else ''
    result += join_tables_horizontally(extracted_jobs['Completed'], extracted_jobs['Failed'])

    if show_all:
        result += '\n\n'
        # TODO pretty handling of empty list in all jobs
        result += extracted_jobs['extra']

    return result


class LsCommand(BaseCommand):
    @staticmethod
    def add_parser(subparsers):
        ls_parser = subparsers.add_parser('ls', help='List all jobs')
        ls_parser.add_argument('-n', '--namespace', help='Namespace to list jobs from', default='All')
        ls_parser.add_argument('-a', '--all', help='Show all jobs', action='store_true', default=False)
        ls_parser.add_argument('-t', '--top', help='Show only first T jobs', default=None, type=int)
        return ls_parser

    def __call__(self, namespace: str = 'All', head: Optional[int] = None, show_all: bool = False):
        jobs = self.backend.list_jobs(namespace=namespace)
        print(visualize_jobs(jobs=jobs, head=head, show_all=show_all))



import time
from collections import defaultdict
from dataclasses import field, dataclass
from enum import Enum
from typing import Literal, Dict, Any, Mapping, Iterable, List, Optional
from datetime import datetime

from kubr.backends.base import BaseBackend
from kubernetes import client, config, watch
from kubernetes.client.models import (  # noqa: F811 redefinition of unused
    V1Container,
    V1ContainerPort,
    V1EmptyDirVolumeSource,
    V1EnvVar,
    V1HostPathVolumeSource,
    V1ObjectMeta,
    V1PersistentVolumeClaimVolumeSource,
    V1Pod,
    V1PodSpec,
    V1ResourceRequirements,
    V1SecurityContext,
    V1Volume,
    V1VolumeMount,
)

from kubr.backends.k8s_runner import (
    create_pod_definition,
)
from kubr.config.runner import RunnerConfig
from kubr.backends.utils import join_tables_horizontally
from kubr.config.job import Job
import humanize
from tabulate import tabulate

class RetryPolicy(str, Enum):
    """
    Defines the retry policy for the ``Roles`` in the ``AppDef``.
    The policy defines the behavior when the role replica encounters a failure:

    1. unsuccessful (non zero) exit code
    2. hardware/host crashes
    3. preemption
    4. eviction

    .. note:: Not all retry policies are supported by all schedulers.
              However all schedulers must support ``RetryPolicy.APPLICATION``.
              Please refer to the scheduler's documentation for more information
              on the retry policies they support and behavior caveats (if any).

    1. REPLICA: Replaces the replica instance. Surviving replicas are untouched.
                Use with ``dist.ddp`` component to have torchelastic coordinate
                restarts and membership changes. Otherwise, it is up to the
                application to deal with failed replica departures and
                replacement replica admittance.
    2. APPLICATION: Restarts the entire application.

    """

    REPLICA = "REPLICA"
    APPLICATION = "APPLICATION"


RETRY_POLICIES: Mapping[str, Iterable[Mapping[str, str]]] = {
    RetryPolicy.REPLICA: [],
    RetryPolicy.APPLICATION: [
        {"event": "PodEvicted", "action": "RestartJob"},
        {"event": "PodFailed", "action": "RestartJob"},
    ],
}


class VolcanoBackend(BaseBackend):
    def __init__(self):
        self.kubernetes_config = config.load_config()
        self.crd_client = client.CustomObjectsApi()
        self.core_client = client.CoreV1Api()

    def run_job(self, run_config: RunnerConfig):
        tasks = []

        for replica_id in range(run_config.resources.num_replicas):
            pod = create_pod_definition(
                pod_name=run_config.experiment.name,
                resource_config=run_config.resources,
                container_config=run_config.container,
                data_config=run_config.data,
                service_account=None,
            )

            # pod.metadata.labels.update(
            #     pod_labels(
            #         app=app,
            #         role_idx=role_idx,
            #         role=role,
            #         replica_id=replica_id,
            #         app_id=unique_app_id,
            #     )
            # )

            task: Dict[str, Any] = {
                "replicas": 1,
                "name": run_config.experiment.task_name,
                "template": pod,
            }
            if run_config.experiment.task_max_retries > 0:
                task["maxRetry"] = run_config.experiment.task_max_retries
                task["policies"] = RETRY_POLICIES[RetryPolicy.APPLICATION]

            if run_config.experiment.min_replicas is not None:
                # first min_replicas tasks are required, afterward optional
                task["minAvailable"] = 1 if replica_id < run_config.experiment.min_replicas else 0

            tasks.append(task)

        job_spec = {
            "schedulerName": "volcano",
            "queue": run_config.experiment.queue,
            "tasks": tasks,
            "maxRetry": run_config.experiment.job_retries,
            "plugins": {
                # https://github.com/volcano-sh/volcano/issues/533
                "svc": ["--publish-not-ready-addresses"],
                "env": [],
            },
        }
        if run_config.experiment.priority_class is not None:
            job_spec["priorityClassName"] = run_config.experiment.priority_class

        resource: Dict[str, object] = {
            "apiVersion": "batch.volcano.sh/v1alpha1",
            "kind": "Job",
            "metadata": {"name": f"{run_config.experiment.name}"},
            "spec": job_spec,
        }

        resp = self.crd_client.create_namespaced_custom_object(
            group="batch.volcano.sh",
            version="v1alpha1",
            namespace=run_config.experiment.namespace,
            plural="jobs",
            body=resource,
        )
        print(resp)

    def _completion_list_running_jobs(self, **kwargs):
        print("Using completion list for running jobs")
        jobs_stat = self.crd_client.list_cluster_custom_object(group='batch.volcano.sh',
                                                               version='v1alpha1',
                                                               plural='jobs')
        jobs = jobs_stat['items']
        running_jobs = []
        for job in jobs:
            if job['status']['state']['phase'] == 'Running':
                running_jobs.append(job['metadata']['name'])
        return running_jobs

    def list_jobs(self, namespace: str = 'All'):
        # TODO [ls] show used resources
        # TODO [ls] speedup for selected namespace(filter on server side)
        # TODO [ls] show events for pending jobs
        jobs_stat = self.crd_client.list_cluster_custom_object(group='batch.volcano.sh',
                                                               version='v1alpha1',
                                                               plural='jobs')
        k8s_jobs = jobs_stat['items']

        extracted_jobs = []
        for k8s_job in k8s_jobs:
            if namespace != 'All' and k8s_job['metadata']['namespace'] != namespace:
                continue
            job = Job(type='Volcano',
                      name=k8s_job['metadata']['name'],
                      namespace=k8s_job['metadata']['namespace'],
                      state=k8s_job['status']['state']['phase'],
                      age=datetime.strptime(k8s_job['status']['state']['lastTransitionTime'],
                                            '%Y-%m-%dT%H:%M:%SZ')
                      )
            extracted_jobs.append(job)

        return extracted_jobs

    def delete_job(self, job_name: str, namespace: str):
        # TODO add cli response formatting for deletion confirmation
        resp = self.crd_client.delete_namespaced_custom_object(group='batch.volcano.sh',
                                                               version='v1alpha1',
                                                               namespace=namespace,
                                                               plural='jobs',
                                                               name=job_name)
        events = self.core_client.list_namespaced_event(namespace=namespace,
                                                        field_selector=f"involvedObject.name={job_name}")
        for event in events.items:
            self.core_client.delete_namespaced_event(name=event.metadata.name, namespace=namespace)

        print(resp)

    def get_job_main_pod(self, job_name: str, namespace: str):
        pods = self.core_client.list_namespaced_pod(namespace=namespace,
                                                    label_selector=f"volcano.sh/job-name={job_name}")
        # TODO add logic for multi pod master-worker selection for logging extraction
        if len(pods.items) == 0:
            raise Exception(f'No pods found for job {job_name} in namespace {namespace}')
        pod = pods.items[0]
        pod_name = pod.metadata.name
        return pod_name, pod

    def get_job_events(self, pod_name: str, namespace: str):
        events = self.core_client.list_namespaced_event(namespace=namespace,
                                                        field_selector=f"involvedObject.name={pod_name}")
        return events

    def get_logs(self, job_name: str, namespace: str, tail: Optional[int] = None):
        pod_name, pod = self.get_job_main_pod(job_name, namespace)
        containers = pod.spec.containers
        if len(containers) == 0:
            return f'No containers found for pod {pod_name} in namespace {namespace}'
        # TODO add logic for multi container selection for logging extraction
        container = containers[0]
        container_name = container.name
        if tail:
            api_response = self.core_client.read_namespaced_pod_log(name=pod_name, namespace=namespace,
                                                                    container=container_name, tail_lines=tail)
            return api_response
        else:
            w = watch.Watch()
            # TODO fix log streaming utf-8 decoding
            for line in w.stream(self.core_client.read_namespaced_pod_log, name=pod_name,
                                 namespace=namespace, container=container_name):
                print(line)

        api_response = self.core_client.read_namespaced_pod_log(name=pod_name, namespace=namespace)
        return api_response

    def describe_job(self, job_name: str, namespace: str):

        pod_name, pod = self.get_job_main_pod(job_name, namespace)
        raw_events = self.get_job_events(pod_name, namespace)

        events = []
        for event in raw_events.items:
            events.append({
                'Last Seen': event.last_timestamp,
                'From': event.source.component,
                'Type': event.type,
                'Reason': event.reason,
                'Message': event.message
            })
        events.sort(key=lambda x: x['Last Seen'], reverse=True)
        for event in events:
            event['Last Seen'] = humanize.naturaldelta(datetime.now() - event['Last Seen'].replace(tzinfo=None))

        # TODO add handling events longer than 10
        events = events[:10]

        events = tabulate(events, headers='keys', tablefmt='grid', maxcolwidths=80)

        return events



from kubr.commands.base import BaseCommand


class StatCommand(BaseCommand):
    @staticmethod
    def add_parser(subparsers, completer):
        stat_parser = subparsers.add_parser('stat', help='Get statistics about a job')
        stat_parser.add_argument('job', help='Name of job to get statistics about').completer = completer
        stat_parser.add_argument('-n', '--namespace', help='Namespace to get statistics from', default='default')
        return stat_parser

    def __call__(self, job_name: str, namespace: str = 'default'):
        raise NotImplementedError



from dataclasses import field
from typing import Dict, List, Literal, Union
from typing import Optional

from pydantic import BaseModel


class ContainerConfig(BaseModel):
    """ ContainerConfig is the configuration for the container.

    Args:
        image (str): Image to run.
        entrypoint (str): Entrypoint to run.
        env (Dict[str, str], optional): Environment variables to pass to the entrypoint. Defaults to {}.

    """
    image: str
    entrypoint: str
    env: Dict[str, str] = field(default_factory=dict)
    port_map: Dict[str, int] = field(default_factory=dict)
    args: List[str] = field(default_factory=list)
    python_path: str = ""


class GitConfig(BaseModel):
    url: str
    branch: str
    commit: str
    secret: str


class VolumeMount(BaseModel):
    """VolumeMount is the configuration for a volume mount.

    Args:
        name (str): Name of the volume.
        type (Literal["pvc", "hostPath"]): Type of the volume.
        mount_path (str): Mount path of the volume.
    """
    name: str
    type: Literal["pvc", "hostPath"]
    mount_path: str


class CodePersistenceConfig(BaseModel):
    git: Optional[GitConfig] = None
    pvc: Optional[str] = None
    volume: Optional[VolumeMount] = None


class DataConfig(BaseModel):
    """DataConfig is the configuration for the data.

    Args:
        pvcs (Optional[List[str]], optional): List of PVCs to mount. Defaults to None.
        volumes (Optional[List[VolumeMount]], optional): List of volumes to mount. Defaults to None.
    """
    pvcs: Optional[List[str]] = None
    volumes: Optional[List[VolumeMount]] = None


class ResourceConfig(BaseModel):
    """ResourceConfig is the configuration for the resources.

    Args:
        num_replicas (int, optional): Number of replicas to run. Defaults to 1.
        cpu (int, optional): Number of CPUs to request. Defaults to 0.
        memMB (int, optional): Memory in MB to request. Defaults to 0.
        gpu (int, optional): Number of GPUs to request. Defaults to 0.
        ib (Union[int, Literal['auto']], optional): Number of Infiniband devices to request. Defaults to 0.
        ib_device (str, optional): Name of the Infiniband device to request. Defaults to "nvidia.com/hostdev".
    """
    num_replicas: int = 1
    cpu: int = 0
    memMB: int = 0
    gpu: int = 0
    devices: Dict[str, float] = field(default_factory=dict)
    capabilities: Dict[str, str] = field(default_factory=dict)
    ib: Union[int, Literal['auto']] = 0
    ib_device: str = "nvidia.com/hostdev"


class ExperimentConfig(BaseModel):
    """ExperimentConfig is the configuration for the experiment.

    Args:
        name (str): Name of the experiment.
        namespace (str): Namespace of the experiment.
        args (List[str], optional): Arguments to pass to the entrypoint. Defaults to [].
        env (Dict[str, str], optional): Environment variables to pass to the entrypoint. Defaults to {}.
        queue (Optional[str], optional): Queue to submit the experiment to. Defaults to "default".
        priority_class (Optional[str], optional): Priority class to submit the experiment to. Defaults to None.
        job_retries (int, optional): Number of retries for the job. Defaults to 0.
        min_replicas (int, optional): Minimum number of replicas to run. Defaults to 1.
        task_name (str, optional): Name of the task. Defaults to "main-task".
        task_max_retries (int, optional): Maximum number of retries for the task. Defaults to 10.
    """
    name: str
    namespace: str

    args: List[str] = field(default_factory=list)
    env: Dict[str, str] = field(default_factory=dict)

    queue: Optional[str] = "default"
    priority_class: Optional[str] = None

    job_retries: int = 0
    min_replicas: int = 1
    task_name: str = "main-task"
    task_max_retries: int = 10


class RunnerConfig(BaseModel):
    """RunnerConfig is the configuration for the runner.

    Args:
        experiment (ExperimentConfig): Experiment configuration.
        container (ContainerConfig): Container configuration.
        resources (ResourceConfig): Resource configuration.
        type (Literal["Volcano"], optional): Type of runner. Defaults to "Volcano".
        code (Optional[CodePersistenceConfig], optional): Code persistence configuration. Defaults to None.
        data (Optional[DataConfig], optional): Data configuration. Defaults to None.

    """
    experiment: ExperimentConfig
    container: ContainerConfig
    resources: ResourceConfig
    type: Literal["Volcano"] = "Volcano"
    code: Optional[CodePersistenceConfig] = None
    data: Optional[DataConfig] = None



from kubr.commands.base import BaseCommand


class AttachCommand(BaseCommand):
    @staticmethod
    def add_parser(subparsers, completer):
        attach_parser = subparsers.add_parser('attach', help='Attach to a running job')
        attach_parser.add_argument('job', help='Name of job to attach to').completer = completer
        attach_parser.add_argument('-n', '--namespace', help='Namespace to attach to', default='default')
        return attach_parser

    def __call__(self, job_name: str, namespace: str = 'default'):
        raise NotImplementedError



from kubr.backends.volcano import VolcanoBackend
from kubr.commands.base import BaseCommand


class DescribeCommand(BaseCommand):
    @staticmethod
    def add_parser(subparsers, completer):
        desc_parser = subparsers.add_parser('desc', help='Get info about a job')
        desc_parser.add_argument('job_name', help='Name of job to get info about').completer = completer
        desc_parser.add_argument('-n', '--namespace', help='Namespace to get info from', default='default')
        return desc_parser

    def __call__(self, *args, **kwargs):
        return self.backend.describe_job(*args, **kwargs)



from pydantic import BaseModel
import datetime
from typing import Literal, Union


class Job(BaseModel):
    type: Literal["Volcano"]
    name: str
    namespace: str
    state: str
    age: Union[datetime.datetime, str]



from typing import Optional

from kubr.backends.volcano import VolcanoBackend


class LogsCommand:
    def __init__(self, backend=None):
        self.backend = backend or VolcanoBackend()

    @staticmethod
    def add_parser(subparsers, completer):
        logs_parser = subparsers.add_parser('logs', help='Get logs of a job')
        logs_parser.add_argument('job',
                                 help='Name of job to get logs of').completer = completer
        logs_parser.add_argument('-n', '--namespace', help='Namespace to get logs from', default='default')
        logs_parser.add_argument('-t', '--tail', help='Number of lines to show', default=10, type=int)
        return logs_parser

    def __call__(self, job_name: str, namespace: str, tail: Optional[int] = None):
        return self.backend.get_logs(job_name=job_name, namespace=namespace, tail=tail)



from dataclasses import dataclass, field
from typing import Dict, List, Literal
from typing import Optional

from kubernetes.client import V1HostPathVolumeSource

from kubr.config.runner import ContainerConfig, ResourceConfig, DataConfig

from kubernetes.client.models import (  # noqa: F811 redefinition of unused
    V1Container,
    V1ContainerPort,
    V1EmptyDirVolumeSource,
    V1EnvVar,
    V1ObjectMeta,
    V1Pod,
    V1PodSpec,
    V1ResourceRequirements,
    V1SecurityContext,
    V1Volume,
    V1VolumeMount,
)

RESERVED_MILLICPU = 100
RESERVED_MEMMB = 1024

ANNOTATION_ISTIO_SIDECAR = "sidecar.istio.io/inject"


def create_pod_definition(pod_name: str, resource_config: ResourceConfig, container_config: ContainerConfig,
                          service_account: Optional[str], data_config: DataConfig) -> "V1Pod":
    limits = {}
    requests = {}

    if resource_config.cpu > 0:
        mcpu = int(resource_config.cpu * 1000)
        limits["cpu"] = f"{mcpu}m"
        request_mcpu = max(mcpu - RESERVED_MILLICPU, 0)
        requests["cpu"] = f"{request_mcpu}m"
    if resource_config.memMB > 0:
        limits["memory"] = f"{int(resource_config.memMB)}M"
        request_memMB = max(int(resource_config.memMB) - RESERVED_MEMMB, 0)
        requests["memory"] = f"{request_memMB}M"
    if resource_config.gpu > 0:
        requests["nvidia.com/gpu"] = limits["nvidia.com/gpu"] = str(resource_config.gpu)
    if resource_config.ib > 0:
        requests[resource_config.ib_device] = limits[resource_config.ib_device] = str(resource_config.ib)

    for device_name, device_limit in resource_config.devices.items():
        limits[device_name] = str(device_limit)

    resources = V1ResourceRequirements(
        limits=limits,
        requests=requests,
    )

    node_selector: Dict[str, str] = {}
    # if LABEL_INSTANCE_TYPE in resource.capabilities:
    #     node_selector[LABEL_INSTANCE_TYPE] = resource.capabilities[LABEL_INSTANCE_TYPE]

    # To support PyTorch dataloaders we need to set /dev/shm to larger than the
    # 64M default so we mount an unlimited sized tmpfs directory on it.
    SHM_VOL = "dshm"
    volumes = [
        V1Volume(
            name=SHM_VOL,
            empty_dir=V1EmptyDirVolumeSource(
                medium="Memory",
            ),
        ),
    ]
    volume_mounts = [
        V1VolumeMount(name=SHM_VOL, mount_path="/dev/shm"),
    ]
    for volume in data_config.volumes:
        if volume.type == "hostPath":
            host_path, mount_path = volume.mount_path.split(":")
            print(volume)
            print(f"Mounting {host_path} at {mount_path}")
            volumes.append(
                V1Volume(
                    name=volume.name,
                    host_path=V1HostPathVolumeSource(
                        path=host_path,
                    )
                )
            )
            volume_mounts.append(
                V1VolumeMount(
                    name=volume.name,
                    mount_path=mount_path,
                )
            )
        else:
            raise ValueError(f"Unknown volume type {volume.type}")
    security_context = V1SecurityContext()

    container = V1Container(
        command=container_config.entrypoint.split(),
        image=container_config.image,
        image_pull_policy="Always",
        name=pod_name,
        env=[
            V1EnvVar(
                name=name,
                value=value,
            )
            for name, value in container_config.env.items()
        ],
        resources=resources,
        ports=[
            V1ContainerPort(
                name=name,
                container_port=port,
            )
            for name, port in container_config.port_map.items()
        ],
        volume_mounts=volume_mounts,
        security_context=security_context,
    )

    return V1Pod(
        spec=V1PodSpec(
            containers=[container],
            restart_policy="Never",
            service_account_name=service_account,
            volumes=volumes,
            node_selector=node_selector,
        ),
        metadata=V1ObjectMeta(
            annotations={
                # Disable the istio sidecar as it prevents the containers from
                # exiting once finished.
                ANNOTATION_ISTIO_SIDECAR: "false",
            },
            labels={},
        ),
    )
